"""Add device_id to clients

Revision ID: 4b93300852aa
Revises: fccd1f95544d
Create Date: 2020-07-11 15:49:38.831459

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4b93300852aa'
down_revision = 'fccd1f95544d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('client', schema=None) as batch_op:
        batch_op.add_column(sa.Column('device_id', sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('client', schema=None) as batch_op:
        batch_op.drop_column('device_id')

    # ### end Alembic commands ###



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2019 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from urllib.parse import quote
from urllib.request import urlopen, Request
from urllib.error import HTTPError
import functools
import json

from colorama import Fore
import click

from ..config import get_token
from ..cliq import cliq

history_count: int = 10

enc = functools.partial(quote, safe="")

friendly_errors = {
    "server_not_found": "Registration target server not found.\n\n"
                        "To log in or register through maubot, you must add the server to the\n"
                        "registration_secrets section in the config. If you only want to log in,\n"
                        "leave the `secret` field empty."
}


@cliq.command(help="Log into a Matrix account via the Maubot server")
@cliq.option("-h", "--homeserver", help="The homeserver to log into", required_unless="list")
@cliq.option("-u", "--username", help="The username to log in with", required_unless="list")
@cliq.option("-p", "--password", help="The password to log in with", inq_type="password",
             required_unless="list")
@cliq.option("-s", "--server", help="The maubot instance to log in through", default="",
             required=False, prompt=False)
@click.option("-r", "--register", help="Register instead of logging in", is_flag=True,
              default=False)
@click.option("-l", "--list", help="List available homeservers", is_flag=True, default=False)
def auth(homeserver: str, username: str, password: str, server: str, register: bool, list: bool
         ) -> None:
    server, token = get_token(server)
    if not token:
        return
    headers = {"Authorization": f"Bearer {token}"}
    if list:
        url = f"{server}/_matrix/maubot/v1/client/auth/servers"
        with urlopen(Request(url, headers=headers)) as resp_data:
            resp = json.load(resp_data)
            print(f"{Fore.GREEN}Available Matrix servers for registration and login:{Fore.RESET}")
            for server in resp.keys():
                print(f"* {Fore.CYAN}{server}{Fore.RESET}")
            return
    endpoint = "register" if register else "login"
    headers["Content-Type"] = "application/json"
    url = f"{server}/_matrix/maubot/v1/client/auth/{enc(homeserver)}/{endpoint}"
    req = Request(url, headers=headers,
                  data=json.dumps({
                      "username": username,
                      "password": password,
                  }).encode("utf-8"))
    try:
        with urlopen(req) as resp_data:
            resp = json.load(resp_data)
            action = "registered" if register else "logged in as"
            print(f"{Fore.GREEN}Successfully {action} "
                  f"{Fore.CYAN}{resp['user_id']}{Fore.GREEN}.")
            print(f"{Fore.GREEN}Access token: {Fore.CYAN}{resp['access_token']}{Fore.RESET}")
            print(f"{Fore.GREEN}Device ID: {Fore.CYAN}{resp['device_id']}{Fore.RESET}")
    except HTTPError as e:
        try:
            err_data = json.load(e)
            error = friendly_errors.get(err_data["errcode"], err_data["error"])
        except (json.JSONDecodeError, KeyError):
            error = str(e)
        action = "register" if register else "log in"
        print(f"{Fore.RED}Failed to {action}: {error}{Fore.RESET}")



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.util.logging.color import (ColorFormatter as BaseColorFormatter, PREFIX, MAU_COLOR,
                                        MXID_COLOR, RESET)

INST_COLOR = PREFIX + "35m"  # magenta
LOADER_COLOR = PREFIX + "36m"  # blue


class ColorFormatter(BaseColorFormatter):
    def _color_name(self, module: str) -> str:
        client = "maubot.client"
        if module.startswith(client):
            return f"{MAU_COLOR}{client}{RESET}.{MXID_COLOR}{module[len(client) + 1:]}{RESET}"
        instance = "maubot.instance"
        if module.startswith(instance):
            return f"{MAU_COLOR}{instance}{RESET}.{INST_COLOR}{module[len(instance) + 1:]}{RESET}"
        loader = "maubot.loader"
        if module.startswith(loader):
            return f"{MAU_COLOR}{instance}{RESET}.{LOADER_COLOR}{module[len(loader) + 1:]}{RESET}"
        if module.startswith("maubot"):
            return f"{MAU_COLOR}{module}{RESET}"
        return super()._color_name(module)



"""Add online field to clients

Revision ID: fccd1f95544d
Revises: d295f8dcfa64
Create Date: 2020-03-06 15:07:50.136644

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'fccd1f95544d'
down_revision = 'd295f8dcfa64'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("client") as batch_op:
        batch_op.add_column(sa.Column('online', sa.Boolean(), nullable=False, server_default=sa.sql.expression.true()))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("client") as batch_op:
        batch_op.drop_column('online')
    # ### end Alembic commands ###



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2019 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union, Awaitable, Optional, Tuple, List
from html import escape
import asyncio

import attr

from mautrix.client import Client as MatrixClient, SyncStream
from mautrix.util.formatter import MatrixParser, MarkdownString, EntityType
from mautrix.util import markdown
from mautrix.types import (EventType, MessageEvent, Event, EventID, RoomID, MessageEventContent,
                           MessageType, TextMessageEventContent, Format, RelatesTo)


class HumanReadableString(MarkdownString):
    def format(self, entity_type: EntityType, **kwargs) -> 'MarkdownString':
        if entity_type == EntityType.URL and kwargs['url'] != self.text:
            self.text = f"{self.text} ({kwargs['url']})"
            return self
        return super(HumanReadableString, self).format(entity_type, **kwargs)


class MaubotHTMLParser(MatrixParser[HumanReadableString]):
    fs = HumanReadableString


def parse_formatted(message: str, allow_html: bool = False, render_markdown: bool = True
                    ) -> Tuple[str, str]:
    if render_markdown:
        html = markdown.render(message, allow_html=allow_html)
    elif allow_html:
        html = message
    else:
        return message, escape(message)
    return MaubotHTMLParser.parse(html).text, html


class MaubotMessageEvent(MessageEvent):
    client: 'MaubotMatrixClient'
    disable_reply: bool

    def __init__(self, base: MessageEvent, client: 'MaubotMatrixClient'):
        super().__init__(**{a.name.lstrip("_"): getattr(base, a.name)
                            for a in attr.fields(MessageEvent)})
        self.client = client
        self.disable_reply = client.disable_replies

    def respond(self, content: Union[str, MessageEventContent],
                event_type: EventType = EventType.ROOM_MESSAGE, markdown: bool = True,
                allow_html: bool = False, reply: Union[bool, str] = False,
                edits: Optional[Union[EventID, MessageEvent]] = None) -> Awaitable[EventID]:
        if isinstance(content, str):
            content = TextMessageEventContent(msgtype=MessageType.NOTICE, body=content)
            if allow_html or markdown:
                content.format = Format.HTML
                content.body, content.formatted_body = parse_formatted(content.body,
                                                                       render_markdown=markdown,
                                                                       allow_html=allow_html)
        if edits:
            content.set_edit(edits)
        elif reply:
            if reply != "force" and self.disable_reply:
                content.body = f"{self.sender}: {content.body}"
                fmt_body = content.formatted_body or escape(content.body).replace("\n", "<br>")
                content.formatted_body = (f'<a href="https://matrix.to/#/{self.sender}">'
                                          f'{self.sender}'
                                          f'</a>: {fmt_body}')
            else:
                content.set_reply(self)
        return self.client.send_message_event(self.room_id, event_type, content)

    def reply(self, content: Union[str, MessageEventContent],
              event_type: EventType = EventType.ROOM_MESSAGE, markdown: bool = True,
              allow_html: bool = False) -> Awaitable[EventID]:
        return self.respond(content, event_type, markdown=markdown, reply=True,
                            allow_html=allow_html)

    def mark_read(self) -> Awaitable[None]:
        return self.client.send_receipt(self.room_id, self.event_id, "m.read")

    def react(self, key: str) -> Awaitable[EventID]:
        return self.client.react(self.room_id, self.event_id, key)

    def edit(self, content: Union[str, MessageEventContent],
             event_type: EventType = EventType.ROOM_MESSAGE, markdown: bool = True,
             allow_html: bool = False) -> Awaitable[EventID]:
        return self.respond(content, event_type, markdown=markdown, edits=self,
                            allow_html=allow_html)


class MaubotMatrixClient(MatrixClient):
    disable_replies: bool

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.disable_replies = False

    def send_markdown(self, room_id: RoomID, markdown: str, *, allow_html: bool = False,
                      msgtype: MessageType = MessageType.TEXT,
                      edits: Optional[Union[EventID, MessageEvent]] = None,
                      relates_to: Optional[RelatesTo] = None, **kwargs
                      ) -> Awaitable[EventID]:
        content = TextMessageEventContent(msgtype=msgtype, format=Format.HTML)
        content.body, content.formatted_body = parse_formatted(markdown, allow_html=allow_html)
        if relates_to:
            if edits:
                raise ValueError("Can't use edits and relates_to at the same time.")
            content.relates_to = relates_to
        elif edits:
            content.set_edit(edits)
        return self.send_message(room_id, content, **kwargs)

    def dispatch_event(self, event: Event, source: SyncStream) -> List[asyncio.Task]:
        if isinstance(event, MessageEvent):
            event = MaubotMessageEvent(event, self)
        elif source != SyncStream.INTERNAL:
            event.client = self
        return super().dispatch_event(event, source)

    async def get_event(self, room_id: RoomID, event_id: EventID) -> Event:
        event = await super().get_event(room_id, event_id)
        if isinstance(event, MessageEvent):
            event.content.trim_reply_fallback()
            return MaubotMessageEvent(event, self)
        else:
            event.client = self
        return event



"""Add Matrix state store

Revision ID: 90aa88820eab
Revises: 4b93300852aa
Create Date: 2020-07-12 01:50:06.215623

"""
from alembic import op
import sqlalchemy as sa

from mautrix.client.state_store.sqlalchemy import SerializableType
from mautrix.types import PowerLevelStateEventContent, RoomEncryptionStateEventContent


# revision identifiers, used by Alembic.
revision = '90aa88820eab'
down_revision = '4b93300852aa'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('mx_room_state',
    sa.Column('room_id', sa.String(length=255), nullable=False),
    sa.Column('is_encrypted', sa.Boolean(), nullable=True),
    sa.Column('has_full_member_list', sa.Boolean(), nullable=True),
    sa.Column('encryption', SerializableType(RoomEncryptionStateEventContent), nullable=True),
    sa.Column('power_levels', SerializableType(PowerLevelStateEventContent), nullable=True),
    sa.PrimaryKeyConstraint('room_id')
    )
    op.create_table('mx_user_profile',
    sa.Column('room_id', sa.String(length=255), nullable=False),
    sa.Column('user_id', sa.String(length=255), nullable=False),
    sa.Column('membership', sa.Enum('JOIN', 'LEAVE', 'INVITE', 'BAN', 'KNOCK', name='membership'), nullable=False),
    sa.Column('displayname', sa.String(), nullable=True),
    sa.Column('avatar_url', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('room_id', 'user_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('mx_user_profile')
    op.drop_table('mx_room_state')
    # ### end Alembic commands ###



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2019 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, Iterable, Optional, Set, Callable, Any, Awaitable, Union, TYPE_CHECKING
from os import path
import asyncio
import logging

from aiohttp import ClientSession
from yarl import URL

from mautrix.errors import MatrixInvalidToken, MatrixRequestError
from mautrix.types import (UserID, SyncToken, FilterID, ContentURI, StrippedStateEvent, Membership,
                           StateEvent, EventType, Filter, RoomFilter, RoomEventFilter, EventFilter,
                           PresenceState, StateFilter)
from mautrix.client import InternalEventType
from mautrix.client.state_store.sqlalchemy import SQLStateStore as BaseSQLStateStore

from .lib.store_proxy import SyncStoreProxy
from .db import DBClient
from .matrix import MaubotMatrixClient

try:
    from mautrix.crypto import (OlmMachine, StateStore as CryptoStateStore, CryptoStore,
                                PickleCryptoStore)


    class SQLStateStore(BaseSQLStateStore, CryptoStateStore):
        pass
except ImportError:
    OlmMachine = CryptoStateStore = CryptoStore = PickleCryptoStore = None
    SQLStateStore = BaseSQLStateStore

try:
    from mautrix.util.async_db import Database as AsyncDatabase
    from mautrix.crypto import PgCryptoStore
except ImportError:
    AsyncDatabase = None
    PgCryptoStore = None

if TYPE_CHECKING:
    from .instance import PluginInstance
    from .config import Config

log = logging.getLogger("maubot.client")


class Client:
    log: logging.Logger = None
    loop: asyncio.AbstractEventLoop = None
    cache: Dict[UserID, 'Client'] = {}
    http_client: ClientSession = None
    global_state_store: Union['BaseSQLStateStore', 'CryptoStateStore'] = SQLStateStore()
    crypto_pickle_dir: str = None
    crypto_db: 'AsyncDatabase' = None

    references: Set['PluginInstance']
    db_instance: DBClient
    client: MaubotMatrixClient
    crypto: Optional['OlmMachine']
    crypto_store: Optional['CryptoStore']
    started: bool

    remote_displayname: Optional[str]
    remote_avatar_url: Optional[ContentURI]

    def __init__(self, db_instance: DBClient) -> None:
        self.db_instance = db_instance
        self.cache[self.id] = self
        self.log = log.getChild(self.id)
        self.references = set()
        self.started = False
        self.sync_ok = True
        self.remote_displayname = None
        self.remote_avatar_url = None
        self.client = MaubotMatrixClient(mxid=self.id, base_url=self.homeserver,
                                         token=self.access_token, client_session=self.http_client,
                                         log=self.log, loop=self.loop, device_id=self.device_id,
                                         sync_store=SyncStoreProxy(self.db_instance),
                                         state_store=self.global_state_store)
        if OlmMachine and self.device_id and (self.crypto_db or self.crypto_pickle_dir):
            self.crypto_store = self._make_crypto_store()
            self.crypto = OlmMachine(self.client, self.crypto_store, self.global_state_store)
            self.client.crypto = self.crypto
        else:
            self.crypto_store = None
            self.crypto = None
        self.client.ignore_initial_sync = True
        self.client.ignore_first_sync = True
        self.client.presence = PresenceState.ONLINE if self.online else PresenceState.OFFLINE
        if self.autojoin:
            self.client.add_event_handler(EventType.ROOM_MEMBER, self._handle_invite)
        self.client.add_event_handler(EventType.ROOM_TOMBSTONE, self._handle_tombstone)
        self.client.add_event_handler(InternalEventType.SYNC_ERRORED, self._set_sync_ok(False))
        self.client.add_event_handler(InternalEventType.SYNC_SUCCESSFUL, self._set_sync_ok(True))

    def _make_crypto_store(self) -> 'CryptoStore':
        if self.crypto_db:
            return PgCryptoStore(account_id=self.id, pickle_key="mau.crypto", db=self.crypto_db)
        elif self.crypto_pickle_dir:
            return PickleCryptoStore(account_id=self.id, pickle_key="maubot.crypto",
                                     path=path.join(self.crypto_pickle_dir, f"{self.id}.pickle"))
        raise ValueError("Crypto database not configured")

    def _set_sync_ok(self, ok: bool) -> Callable[[Dict[str, Any]], Awaitable[None]]:
        async def handler(data: Dict[str, Any]) -> None:
            self.sync_ok = ok

        return handler

    async def start(self, try_n: Optional[int] = 0) -> None:
        try:
            if try_n > 0:
                await asyncio.sleep(try_n * 10)
            await self._start(try_n)
        except Exception:
            self.log.exception("Failed to start")

    async def _start(self, try_n: Optional[int] = 0) -> None:
        if not self.enabled:
            self.log.debug("Not starting disabled client")
            return
        elif self.started:
            self.log.warning("Ignoring start() call to started client")
            return
        try:
            user_id = await self.client.whoami()
        except MatrixInvalidToken as e:
            self.log.error(f"Invalid token: {e}. Disabling client")
            self.db_instance.enabled = False
            return
        except MatrixRequestError:
            if try_n >= 5:
                self.log.exception("Failed to get /account/whoami, disabling client")
                self.db_instance.enabled = False
            else:
                self.log.exception(f"Failed to get /account/whoami, "
                                   f"retrying in {(try_n + 1) * 10}s")
                _ = asyncio.ensure_future(self.start(try_n + 1), loop=self.loop)
            return
        if user_id != self.id:
            self.log.error(f"User ID mismatch: expected {self.id}, but got {user_id}")
            self.db_instance.enabled = False
            return
        if not self.filter_id:
            self.db_instance.edit(filter_id=await self.client.create_filter(Filter(
                room=RoomFilter(
                    timeline=RoomEventFilter(
                        limit=50,
                        lazy_load_members=True,
                    ),
                    state=StateFilter(
                        lazy_load_members=True,
                    )
                ),
                presence=EventFilter(
                    not_types=[EventType.PRESENCE],
                ),
            )))
        if self.displayname != "disable":
            await self.client.set_displayname(self.displayname)
        if self.avatar_url != "disable":
            await self.client.set_avatar_url(self.avatar_url)
        if self.crypto:
            self.log.debug("Enabling end-to-end encryption support")
            await self.crypto_store.open()
            crypto_device_id = await self.crypto_store.get_device_id()
            if crypto_device_id and crypto_device_id != self.device_id:
                self.log.warning("Mismatching device ID in crypto store and main database. "
                                 "Encryption may not work.")
            await self.crypto.load()
            if not crypto_device_id:
                await self.crypto_store.put_device_id(self.device_id)
        self.start_sync()
        await self._update_remote_profile()
        self.started = True
        self.log.info("Client started, starting plugin instances...")
        await self.start_plugins()

    async def start_plugins(self) -> None:
        await asyncio.gather(*[plugin.start() for plugin in self.references])

    async def stop_plugins(self) -> None:
        await asyncio.gather(*[plugin.stop() for plugin in self.references if plugin.started])

    def start_sync(self) -> None:
        if self.sync:
            self.client.start(self.filter_id)

    def stop_sync(self) -> None:
        self.client.stop()

    async def stop(self) -> None:
        if self.started:
            self.started = False
            await self.stop_plugins()
            self.stop_sync()
            if self.crypto:
                await self.crypto_store.close()

    def clear_cache(self) -> None:
        self.stop_sync()
        self.db_instance.edit(filter_id="", next_batch="")
        self.start_sync()

    def delete(self) -> None:
        try:
            del self.cache[self.id]
        except KeyError:
            pass
        self.db_instance.delete()

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "homeserver": self.homeserver,
            "access_token": self.access_token,
            "device_id": self.device_id,
            "enabled": self.enabled,
            "started": self.started,
            "sync": self.sync,
            "sync_ok": self.sync_ok,
            "autojoin": self.autojoin,
            "online": self.online,
            "displayname": self.displayname,
            "avatar_url": self.avatar_url,
            "remote_displayname": self.remote_displayname,
            "remote_avatar_url": self.remote_avatar_url,
            "instances": [instance.to_dict() for instance in self.references],
        }

    @classmethod
    def get(cls, user_id: UserID, db_instance: Optional[DBClient] = None) -> Optional['Client']:
        try:
            return cls.cache[user_id]
        except KeyError:
            db_instance = db_instance or DBClient.get(user_id)
            if not db_instance:
                return None
            return Client(db_instance)

    @classmethod
    def all(cls) -> Iterable['Client']:
        return (cls.get(user.id, user) for user in DBClient.all())

    async def _handle_tombstone(self, evt: StateEvent) -> None:
        if not evt.content.replacement_room:
            self.log.info(f"{evt.room_id} tombstoned with no replacement, ignoring")
            return
        _, server = self.client.parse_user_id(evt.sender)
        await self.client.join_room(evt.content.replacement_room, servers=[server])

    async def _handle_invite(self, evt: StrippedStateEvent) -> None:
        if evt.state_key == self.id and evt.content.membership == Membership.INVITE:
            await self.client.join_room(evt.room_id)

    async def update_started(self, started: bool) -> None:
        if started is None or started == self.started:
            return
        if started:
            await self.start()
        else:
            await self.stop()

    async def update_displayname(self, displayname: str) -> None:
        if displayname is None or displayname == self.displayname:
            return
        self.db_instance.displayname = displayname
        if self.displayname != "disable":
            await self.client.set_displayname(self.displayname)
        else:
            await self._update_remote_profile()

    async def update_avatar_url(self, avatar_url: ContentURI) -> None:
        if avatar_url is None or avatar_url == self.avatar_url:
            return
        self.db_instance.avatar_url = avatar_url
        if self.avatar_url != "disable":
            await self.client.set_avatar_url(self.avatar_url)
        else:
            await self._update_remote_profile()

    async def update_access_details(self, access_token: str, homeserver: str) -> None:
        if not access_token and not homeserver:
            return
        elif access_token == self.access_token and homeserver == self.homeserver:
            return
        new_client = MaubotMatrixClient(mxid=self.id, base_url=homeserver or self.homeserver,
                                        token=access_token or self.access_token, loop=self.loop,
                                        client_session=self.http_client, device_id=self.device_id,
                                        log=self.log, state_store=self.global_state_store)
        mxid = await new_client.whoami()
        if mxid != self.id:
            raise ValueError(f"MXID mismatch: {mxid}")
        new_client.sync_store = SyncStoreProxy(self.db_instance)
        self.stop_sync()
        self.client = new_client
        self.db_instance.homeserver = homeserver
        self.db_instance.access_token = access_token
        self.start_sync()

    async def _update_remote_profile(self) -> None:
        profile = await self.client.get_profile(self.id)
        self.remote_displayname, self.remote_avatar_url = profile.displayname, profile.avatar_url

    # region Properties

    @property
    def id(self) -> UserID:
        return self.db_instance.id

    @property
    def homeserver(self) -> str:
        return self.db_instance.homeserver

    @property
    def access_token(self) -> str:
        return self.db_instance.access_token

    @property
    def device_id(self) -> str:
        return self.db_instance.device_id

    @property
    def enabled(self) -> bool:
        return self.db_instance.enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.db_instance.enabled = value

    @property
    def next_batch(self) -> SyncToken:
        return self.db_instance.next_batch

    @property
    def filter_id(self) -> FilterID:
        return self.db_instance.filter_id

    @property
    def sync(self) -> bool:
        return self.db_instance.sync

    @sync.setter
    def sync(self, value: bool) -> None:
        if value == self.db_instance.sync:
            return
        self.db_instance.sync = value
        if self.started:
            if value:
                self.start_sync()
            else:
                self.stop_sync()

    @property
    def autojoin(self) -> bool:
        return self.db_instance.autojoin

    @autojoin.setter
    def autojoin(self, value: bool) -> None:
        if value == self.db_instance.autojoin:
            return
        if value:
            self.client.add_event_handler(EventType.ROOM_MEMBER, self._handle_invite)
        else:
            self.client.remove_event_handler(EventType.ROOM_MEMBER, self._handle_invite)
        self.db_instance.autojoin = value

    @property
    def online(self) -> bool:
        return self.db_instance.online

    @online.setter
    def online(self, value: bool) -> None:
        self.client.presence = PresenceState.ONLINE if value else PresenceState.OFFLINE
        self.db_instance.online = value

    @property
    def displayname(self) -> str:
        return self.db_instance.displayname

    @property
    def avatar_url(self) -> ContentURI:
        return self.db_instance.avatar_url

    # endregion


def init(config: 'Config', loop: asyncio.AbstractEventLoop) -> Iterable[Client]:
    Client.http_client = ClientSession(loop=loop)
    Client.loop = loop

    if OlmMachine:
        db_type = config["crypto_database.type"]
        if db_type == "default":
            db_url = config["database"]
            parsed_url = URL(db_url)
            if parsed_url.scheme == "sqlite":
                Client.crypto_pickle_dir = config["crypto_database.pickle_dir"]
            elif parsed_url.scheme == "postgres":
                if not PgCryptoStore:
                    log.warning("Default database is postgres, but asyncpg is not installed. "
                                "Encryption will not work.")
                else:
                    Client.crypto_db = AsyncDatabase(url=db_url,
                                                     upgrade_table=PgCryptoStore.upgrade_table)
        elif db_type == "pickle":
            Client.crypto_pickle_dir = config["crypto_database.pickle_dir"]
        elif db_type == "postgres" and PgCryptoStore:
            Client.crypto_db = AsyncDatabase(url=config["crypto_database.postgres_uri"],
                                             upgrade_table=PgCryptoStore.upgrade_table)
        else:
            raise ValueError("Unsupported crypto database type")

    return Client.all()
